# Importación de librerias
import pandas as pd
import numpy as np
import datetime
import logging
logging.getLogger('tensorflow').disabled = True
import tensorflow as tf
import pandas.api.types as ptypes
from sklearn.metrics import confusion_matrix

# Control de tiempo
start_time = datetime.datetime.now()
print("Inicio del proceso: " + start_time.strftime("%d/%m/%Y %H:%M:%S"))

print("***** INICIO DE LA FASE DE IMPORTACIÓN DE DATOS *****")
# Importación de los datos para el entrenamiento
print("Importando los datos para de la fase de entrenamiento...")
df_train = pd.read_csv('../validation_data_out.csv')
print("Preparando dataframe con las columnas de características...")
x_train = df_train.iloc[ : , 3:]
print("Preparando dataframe con la columna de etiquetas...")
y_train = df_train.iloc[ : , 0:1]
print("Filas:" + str(len(x_train)))
print("Columnas:" + str(len(x_train.columns)))

# Importación de los datos a clasificar
print("Importando los datos para la fase de clasificación predictiva...")
df_pred = pd.read_csv('../data_to_predict_out.csv')
print("Preparando dataframe con las columnas de características...")
x_pred = df_pred.iloc[ : , 2:]
print("Preparando dataframe con la columna de etiquetas...")
y_pred = df_pred.iloc[ : , 0:1]
print("Filas:" + str(len(x_pred)))
print("Columnas:" + str(len(x_pred.columns)))
print("***** FIN DE LA FASE DE IMPORTACIÓN DE DATOS *****")

print("***** INICIO DE LA FASE DE DEFINICIÓN DE PARÁMETROS DE AJUSTE Y MODELO *****")
# Hiperparámetros
train_num_epochs=10 # Ajustar en orden ascendente
validation_num_epochs=1 # Ajustar en orden ascendente
steps=25000 # Ajustar en orden ascendente. max(steps) = (len(x_train) / batch_size(default:128)) * train_num_epochs
learning_rate = 0.01 # Entre 0 y 1. Ajustar en orden descendente
# Cantidad de neuronas en las capas ocultas calculadas según: (num. características de entrada + num. clases de salida) * (2/3)
# (41+2)*(2/3) = ~28. Redondeado a potencia de 2 => 32
n_hidden_1 = 32 # Número de neuronas de la primera capa oculta
n_hidden_2 = 32 # Número de neuronas de la segunda capa oculta

# Función de entrada para alimentar el modelo en la fase de predicción
pred_input_fn = tf.estimator.inputs.pandas_input_fn(
	x=x_pred,
	y=y_pred,
	shuffle=False)

# Transformación de las columnas de características
feature_columns = []
for col in x_train.columns:
	if ptypes.is_string_dtype(x_train[col]): # Para las columnas con valores categóricos
		# hash_bucket_size = número de categorías diferentes en la columna
		# dimension = raíz cuarta del número de categorías diferentes en la columna
		feature_columns.append(tf.feature_column.embedding_column(
				tf.feature_column.categorical_column_with_hash_bucket(
					col, hash_bucket_size=len(x_train[col].unique())),
						dimension=round(len(x_train[col].unique())**0.25)))

	elif ptypes.is_numeric_dtype(x_train[col]): # Para las columnas con valores numéricos
		feature_columns.append(tf.feature_column.numeric_column(col))

# Definición del modelo clasificador de red neural profunda
modelo = tf.estimator.DNNClassifier(
        hidden_units=[n_hidden_1, n_hidden_2],
	feature_columns=feature_columns,
	model_dir='./mod9',
	n_classes=2,
	label_vocabulary=["NORMAL", "ATTACK"],
	optimizer=tf.train.ProximalAdagradOptimizer(
		learning_rate=learning_rate,
		l1_regularization_strength=0.001))
print("***** FIN DE LA FASE DE DEFINICIÓN DE PARÁMETROS DE AJUSTE Y MODELO *****")

# Fase de clasificación predictiva
print("***** INICIO DE LA FASE DE CLASIFICACIÓN PREDICTIVA *****")
predicted_results = modelo.predict(input_fn=pred_input_fn)
print("***** FIN DE LA FASE DE CLASIFICACIÓN PREDICTIVA *****")

# Resultados de precisión del algoritmo
print("***** EXPORTACIÓN DE RESULTADOS A FICHERO CSV *****")
prediction_ids = [prediction['class_ids'][0] for prediction in predicted_results]
predicted = pd.DataFrame({"label_class": prediction_ids})
predicted.replace({0: 'NORMAL', 1: 'ATTACK'}, inplace=True)
predicted_out = pd.concat([x_pred, predicted, df_pred.iloc[ : , 1:2]], axis=1)
predicted_out.to_csv('../predicted_data.csv', sep=',', header=None, index=False)

# Control de tiempo.
finish_time = datetime.datetime.now()
print("Fin del proceso: " + finish_time.strftime("%d/%m/%Y %H:%M:%S")) # Fin del proceso
total_time = finish_time - start_time
print("Tiempo total empleado en la ejecución: " + str(total_time)) # Tiempo total empleado
print("***** FIN *****")
