# Importación de librerias
import pandas as pd
import numpy as np
import datetime
import logging
logging.getLogger('tensorflow').disabled = True
import tensorflow as tf
import pandas.api.types as ptypes
from sklearn.metrics import confusion_matrix

# Control de tiempo
start_time = datetime.datetime.now()
print("Inicio del proceso: " + start_time.strftime("%d/%m/%Y %H:%M:%S"))

print("***** INICIO DE LA FASE DE IMPORTACIÓN DE DATOS *****")
# Importación de los datos para el entrenamiento
print("Importando los datos para la fase de entrenamiento...")
df_train = pd.read_csv('../validation_data_out.csv')
print("Preparando dataframe con las columnas de características...")
x_train = df_train.iloc[ : , 3:]
print("Preparando dataframe con la columna de etiquetas...")
y_train = df_train.iloc[ : , 0:1]
print("Filas:" + str(len(x_train)))
print("Columnas:" + str(len(x_train.columns)))
# Control de tiempo. Importacion del fichero de datos de entrenamiento
train_data_import_time = datetime.datetime.now()

# Importación de los datos para la fase de validación
print("Importando los datos para la fase de validación...")
df_validation = df_train.sample(frac=0.2)
print("Preparando dataframe con las columnas de características...")
x_validation = df_validation.iloc[ : , 3:]
print("Preparando dataframe con la columna de etiquetas...")
y_validation = df_validation.iloc[ : , 0:1]
print("Filas:" + str(len(x_validation)))
print("Columnas:" + str(len(x_validation.columns)))
# Control de tiempo. Importacion del fichero de datos de validacion
validation_data_import_time = datetime.datetime.now()

# Importación de los datos para el test
print("Importando los datos para la fase de test...")
print("Preparando dataframe con las columnas de características...")
df_test = pd.read_csv('../test_data_out.csv')
x_test = df_test.iloc[ : , 3:]
print("Preparando dataframe con la columna de etiquetas...")
y_test = df_test.iloc[ : , 0:1]
print("Filas:" + str(len(x_test)))
print("Columnas:" + str(len(x_test.columns)))
# Control de tiempo. Importacion del fichero de datos de test
test_data_import_time = datetime.datetime.now()
print("***** FIN DE LA FASE DE IMPORTACIÓN DE DATOS *****")

print("***** INICIO DE LA FASE DE DEFINICIÓN DE PARÁMETROS DE AJUSTE Y MODELO *****")
# Hiperparámetros
train_num_epochs=10 # Ajustar en orden ascendente
validation_num_epochs=1 # Ajustar en orden ascendente
steps=25000 # Ajustar en orden ascendente. max(steps) = (len(x_train) / batch_size(default:128)) * train_num_epochs
learning_rate = 0.01 # Entre 0 y 1. Ajustar en orden descendente
# Cantidad de neuronas en las capas ocultas calculadas según: (num. características de entrada + num. clases de salida) * (2/3)
# (41+2)*(2/3) = ~28. Redondeado a potencia de 2 => 32
n_hidden_1 = 32 # Número de neuronas de la primera capa oculta
n_hidden_2 = 32 # Número de neuronas de la segunda capa oculta

# Función de entrada para alimentar el modelo en la fase de entrenamiento
train_input_fn = tf.estimator.inputs.pandas_input_fn(
	x=x_train,
	y=y_train,
	num_epochs=train_num_epochs,
	shuffle=True)

# Función de entrada para alimentar el modelo en la fase de validación
validation_input_fn = tf.estimator.inputs.pandas_input_fn(
	x=x_validation,
	y=y_validation,
	num_epochs=validation_num_epochs,
	shuffle=False)

# Función de entrada para alimentar el modelo en la fase de predicción
test_input_fn = tf.estimator.inputs.pandas_input_fn(
	x=x_test,
	y=y_test,
	shuffle=False)

# Transformación de las columnas de características
feature_columns = []
for col in x_train.columns:
	if ptypes.is_string_dtype(x_train[col]): # Para las columnas con valores categóricos
		# hash_bucket_size = número de categorías diferentes en la columna
		# dimension = raíz cuarta del número de categorías diferentes en la columna
		feature_columns.append(tf.feature_column.embedding_column(
				tf.feature_column.categorical_column_with_hash_bucket(
					col, hash_bucket_size=len(x_train[col].unique())),
						dimension=round(len(x_train[col].unique())**0.25)))

	elif ptypes.is_numeric_dtype(x_train[col]): # Para las columnas con valores numéricos
		feature_columns.append(tf.feature_column.numeric_column(col))

# Definición del modelo clasificador de red neural profunda
modelo = tf.estimator.DNNClassifier(
        hidden_units=[n_hidden_1, n_hidden_2],
	feature_columns=feature_columns,
	model_dir='./mod9',
	n_classes=2,
	label_vocabulary=["NORMAL", "ATTACK"],
	optimizer=tf.train.ProximalAdagradOptimizer(
		learning_rate=learning_rate,
		l1_regularization_strength=0.001))
print("***** FIN DE LA FASE DE DEFINICIÓN DE PARÁMETROS DE AJUSTE Y MODELO *****")

# Fase de entrenamiento
print("***** INICIO DE LA FASE DE ENTRENAMIENTO *****")
modelo.train(input_fn=train_input_fn);
print("***** FIN DE LA FASE DE ENTRENAMIENTO *****")
# Control de tiempo de la fase de entrenamiento
train_time = datetime.datetime.now()

# Fase de validación con los datos de validación
print("***** INICIO DE LA FASE DE VALIDACIÓN *****")
validation_eval_result = modelo.evaluate(input_fn=validation_input_fn)
print("***** FIN DE LA FASE DE VALIDACIÓN *****")
# Control de tiempo de la fase de validación
validation_time = datetime.datetime.now()

# Fase de validación predicción con los datos de test
print("***** INICIO DE LA FASE DE TEST *****")
test_eval_result = modelo.evaluate(input_fn=test_input_fn)
predict_results = modelo.predict(input_fn=test_input_fn)
print("***** FIN DE LA FASE DE TEST *****")
# Control de tiempo de la fase de predicción
test_time = datetime.datetime.now()

# Tiempos
print("***** TIEMPOS *****")
# Inicio del proceso
print("Inicio del proceso: " + start_time.strftime("%d/%m/%Y %H:%M:%S")) # Tiempo empleado en la importación de datos de entrenamiento
print("Tiempo empleado en la importación de datos de entrenamiento: " + 
		str(train_data_import_time - start_time)) # Tiempo empleado en la importación de datos de validation
print("Tiempo empleado en la importación de datos de validación: " + 
		str(validation_data_import_time - train_data_import_time)) # Tiempo empleado en la importación de datos de test
print("Tiempo empleado en la importación de datos de test: " + 
		str(test_data_import_time - validation_data_import_time)) # Tiempo empleado en la fase de entrenamiento
print("Tiempo empleado en la fase de entrenamiento: " + 
		str(train_time - test_data_import_time)) # Tiempo empleado en la fase de validación
print("Tiempo empleado en la fase de validación: " + 
		str(validation_time - train_time)) # Tiempo empleado en la fase de predicción
print("Tiempo empleado en la fase de predicción: " + 
		str(test_time-validation_time))

# Resultados de precisión del algoritmo
print("***** PRECISIÓN *****")
test_labels = y_test
test_labels.replace({'NORMAL': 0, 'ATTACK': 1}, inplace=True)
test_labels = np.asarray(test_labels, np.float32)
prediction_ids = [prediction['class_ids'][0] for prediction in predict_results]
predict = pd.DataFrame({"label_class": prediction_ids})
predicted_labels = np.asarray(predict, np.float32)
vn, fp, fn, vp = confusion_matrix(test_labels, predicted_labels).ravel()
print('VN= ' + str(vn) + ', FP = ' + str(fp) + ', FN = ' + str(fn) + ', VP = ' + str(vp))
vpp = (vp / (vp + fn)) * 100
vnp = (vn / (fp + vn)) * 100
sen = (vp / (vp + fp)) * 100
esp = (vn / (fn + vn)) * 100
print("Valor de predicción positivo: " + str(vpp))
print("Valor de predicción negativo: " + str(vnp))
print("Sensibilidad: " + str(sen))
print("Especificidad: " + str(esp))
print("Error Tipo I: " + str(fp))
print("Error Tipo II: " + str(fn))
print ("Precisión en la validación: {accuracy}".format(**validation_eval_result))
print ("Precisión en el test: {accuracy}".format(**test_eval_result))

# Control de tiempo.
finish_time = datetime.datetime.now()
print("Fin del proceso: " + finish_time.strftime("%d/%m/%Y %H:%M:%S")) # Fin del proceso
total_time = finish_time - start_time
print("Tiempo total empleado en la ejecución: " + str(total_time)) # Tiempo total empleado
print("***** FIN *****")